import colorsys
from decimal import Decimal

from . import exceptions, helpers
from .typing import ColorType


class Color:
    def __init__(self, color: ColorType):
        self._rgba = helpers.format_color(color)

    def __str__(self):
        if self.opaque:
            return self.hex()
        else:
            return self.rgba()

    def __repr__(self):
        return "%s('%s')" % (
            self.__class__.__name__,
            self.hex() if self.opaque else self.hexa(),
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._rgba == other._rgba

        try:
            other_color = self.__class__(other)
        except exceptions.InvalidColor:
            return super().__eq__(other)
        else:
            return self._rgba == other_color._rgba

    def __iter__(self):
        return iter(self._rgba)

    def __getitem__(self, item):
        return self._rgba[item]

    @property
    def opaque(self):
        return self._rgba[3] == 255

    def hex(self, value=None):
        if value is None:
            return '#{:02X}{:02X}{:02X}'.format(*self._rgba[:3])
        else:
            color_bytes = helpers.format_hexa(value)
            color_bytes_list = list(color_bytes)
            color_bytes_list[3] = self._rgba[3]  # keep current opacity
            return type(self)(color_bytes_list)

    def hexa(self, value=None):
        if value is None:
            return '#{:02X}{:02X}{:02X}{:02X}'.format(*self._rgba)
        else:
            return type(self)(helpers.format_hexa(value))

    def opacity(self, value=None):
        if value is None:
            # strip trailing zeros
            return round(Decimal(self._rgba[3]) * 100 / 255) / 100
        else:
            opacity = round(Decimal(str(value)) * 255)
            if not 0 <= opacity <= 255:
                raise exceptions.InvalidOpacity(value)

            color_bytes = list(self._rgba[:3])
            color_bytes.append(opacity)
            return type(self)(color_bytes)

    def rgb(self):
        return "rgb({:d}, {:d}, {:d})".format(*self._rgba[:3])

    def rgba(self):
        return "rgba({:d}, {:d}, {:d}, {})".format(
            self._rgba[0], self._rgba[1], self._rgba[2], self.opacity()
        )

    def hsl(self):
        h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in self._rgba[:3]))
        return "hsl({:d}, {}%, {}%)".format(
            round(h * 360), round(s * 1000) / 10, round(l * 1000) / 10
        )

    def hsla(self):
        h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in self._rgba[:3]))
        return "hsla({:d}, {}%, {}%, {})".format(
            round(h * 360), round(s * 1000) / 10, round(l * 1000) / 10, self.opacity()
        )



from django import forms
from django.core import exceptions
from django.utils.translation import gettext_lazy as _

from .color import Color
from .exceptions import InvalidColor
from .widgets import ColorWidget


class ColorField(forms.CharField):
    widget = ColorWidget
    default_error_messages = {
        'invalid': _('Enter a valid color.'),
    }

    def clean(self, value):
        if value in self.empty_values:
            return self.empty_value

        try:
            return Color(value)
        except InvalidColor:
            raise exceptions.ValidationError(
                self.error_messages['invalid'], code='invalid'
            )



import re
from decimal import Decimal
from typing import Iterable

from . import exceptions
from .typing import ColorBytes, ColorType

re_hexa = re.compile(r'#?([\dA-Fa-f]{3,4}|[\dA-Fa-f]{6}|[\dA-Fa-f]{8})')
re_rgba = re.compile(
    r'rgba?[\D]+?([-\d]+)[\D]+?([-\d]+)[\D]+?([-\d]+)(?:[\D]+?([-\d.]+))?[\D]+'
)


def format_color_byte(value) -> int:
    if isinstance(value, int):
        component = value
    elif isinstance(value, str):
        component = int(value)
    else:
        raise TypeError(value)

    if not 0 <= component <= 255:
        raise ValueError

    return component


def format_color_bytes(sequence) -> ColorBytes:
    """
    Examples:
      ["128", "192", "128"] => (128, 192, 128, 255)
      ["128", "192", "0", "64"] => (128, 192, 0, 64)
    """
    components = list(sequence)
    if len(components) not in {3, 4}:
        raise ValueError(components)

    if len(components) == 3:
        components.append(255)

    return tuple(map(format_color_byte, components))


def format_rgba(sequence) -> ColorBytes:
    """
    Examples:
      [128, 192, 64] => (128, 192, 64, 255)
      ["128", "192", "64", "0.5"] => (128, 192, 64, 128)
    """
    components = list(sequence)

    # Convert fraction to byte
    # https://stackoverflow.com/questions/5445085/understanding-colors-on-android-six-characters/11019879#11019879
    if len(components) == 4:
        opacity = components[3]

        if not isinstance(opacity, str):
            opacity = str(opacity)

        components[3] = round(Decimal(opacity) * 255)

    return format_color_bytes(components)


def format_hexa(value: str) -> ColorBytes:
    """
    Examples:
      "bda" => (187, 221, 170, 255)
      "4fcd" => (68, 255, 204, 221)
      "60B0C4" => (96, 176, 196, 255)
      "2BEA40D0" => (43, 234, 64, 208)
    """
    if len(value) in {3, 4}:
        expanded_color = ''.join(s * 2 for s in value)
    else:
        expanded_color = value

    length = len(expanded_color)
    if length in {6, 8}:
        hex_parts = [expanded_color[i:(i + 2)] for i in range(0, length, 2)]
        return format_color_bytes([int(v, 16) for v in hex_parts])
    else:
        raise ValueError(value)


def format_color(value: ColorType) -> ColorBytes:
    if isinstance(value, str):
        hexa_match = re_hexa.fullmatch(value)
        if hexa_match is not None:
            try:
                return format_hexa(hexa_match.group(1))
            except (ValueError, TypeError):
                raise exceptions.InvalidColorValue(value)

        rgba_match = re_rgba.fullmatch(value)
        if rgba_match is not None:
            color_items = rgba_match.groups()
            if color_items[3] is None:
                color_items = color_items[:3]

            try:
                return format_rgba(color_items)
            except (ValueError, TypeError):
                raise exceptions.InvalidColorValue(value)

        raise exceptions.InvalidColorValue(value)
    elif isinstance(value, Iterable):
        try:
            return format_color_bytes(value)
        except (ValueError, TypeError):
            raise exceptions.InvalidColorValue(value)
    else:
        raise exceptions.InvalidColorType(value)



from django import forms


class ColorWidget(forms.HiddenInput):
    template_name = "spectrum/widget.html"

    @property
    def is_hidden(self):
        return False

    @property
    def media(self):
        return forms.Media(
            css={
                "screen": ["spectrum/css/monolith.min.css", "spectrum/css/widget.css"]
            },
            js=["spectrum/js/pickr.es5.min.js", "spectrum/js/widget.js"],
        )



from decimal import Decimal

from django.core import exceptions
from django.db import models
from django.utils.translation import gettext_lazy as _

from . import forms
from .color import Color
from .exceptions import InvalidColor


def encode_color(color):
    return color.hex() if color.opaque else color.hexa()


def parse_color(value: str):
    if ":" in value:
        # v0.1.0 compat
        color, _, opacity = value.partition(":")
        hex_opacity = "{:02X}".format(round(Decimal(str(opacity)) * 255))
        return Color(color + hex_opacity)
    else:
        return Color(value)


class ColorField(models.Field):
    default_error_messages = {
        "invalid": _("'%(value)s' value must be a RGB/RGBA color."),
    }
    description = _("RGB/RGBA color")

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 12
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "CharField"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.max_length == 12:
            del kwargs["max_length"]
        return name, path, args, kwargs

    def from_db_value(self, value, *args, **kwargs):
        if value in self.empty_values:
            return None
        return parse_color(value)

    def to_python(self, value):
        if isinstance(value, Color):
            return value

        if value in self.empty_values:
            return None

        try:
            return parse_color(value)
        except InvalidColor:
            raise exceptions.ValidationError(
                self.error_messages["invalid"], code="invalid", params={"value": value},
            )

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        if value in self.empty_values:
            return ""
        return encode_color(value)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_prep_value(value)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': forms.ColorField,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)
